"""This module brings in-place mutations"""
from .base import BaseMutation


class BaseNameValueMutation(BaseMutation):
    """A base in-place mutation where name and value are needed to define it"""

    def __init__(self, name, value):
        """
        Args:
          name: to be wrapped with :py:obj:`ensure_conversion` and used as
            a key/attr/index for a mutation
          value: to be wrapped with :py:obj:`ensure_conversion` and used as
            a value for a mutation
        """
        super().__init__()
        self.name = self.ensure_conversion(name)
        self.value = self.ensure_conversion(value)


class BaseIndexMutation(BaseMutation):
    def __init__(self, index):
        super().__init__()
        self.index = self.ensure_conversion(index)


class SetItem(BaseNameValueMutation):
    def _gen_code_and_update_ctx(self, code_input, ctx):
        name_code = self.name.gen_code_and_update_ctx(code_input, ctx)
        value_code = self.value.gen_code_and_update_ctx(code_input, ctx)
        return f"{code_input}[{name_code}] = {value_code}"


class SetAttr(BaseNameValueMutation):
    def _gen_code_and_update_ctx(self, code_input, ctx):
        name_code = self.name.gen_code_and_update_ctx(code_input, ctx)
        value_code = self.value.gen_code_and_update_ctx(code_input, ctx)
        return f"setattr({code_input}, {name_code}, {value_code})"


class DelItem(BaseIndexMutation):
    def _gen_code_and_update_ctx(self, code_input, ctx):
        index_code = self.index.gen_code_and_update_ctx(code_input, ctx)
        return f"{code_input}.pop({index_code})"


class DelAttr(BaseIndexMutation):
    def _gen_code_and_update_ctx(self, code_input, ctx):
        index_code = self.index.gen_code_and_update_ctx(code_input, ctx)
        return f"delattr({code_input}, {index_code})"


class Custom(BaseMutation):
    """Mutation to be used in conjunction with `tap` method.
    Runs the code, defined by the conversion argument and returns the input
    as is."""

    def __init__(self, conversion):
        """
        Arg:
          conversion: to be wrapped with :py:obj:`ensure_conversion` and used
            as a mutation code
        """
        super().__init__()
        self.conversion = self.ensure_conversion(conversion)

    def _gen_code_and_update_ctx(self, code_input, ctx):
        return self.conversion.gen_code_and_update_ctx(code_input, ctx)


class Mutations:
    """Mutations to be used in conjunction with the conversion `tap` method."""

    #: Sets a value by key.
    set_item = SetItem
    #: Sets a value of an attribute.
    set_attr = SetAttr
    #: Pops a key from a dict.
    del_item = DelItem
    #: Deletes an attribute from an object.
    del_attr = DelAttr
    #: Runs the code, defined by the conversion argument and returns the input
    #: as is.
    custom = Custom



"""
This module brings join functionality to the library
"""
import typing as t
from itertools import chain

from .aggregations import Aggregate, ReduceFuncs
from .base import (
    And,
    BaseConversion,
    CallFunc,
    ConversionWrapper,
    Eq,
    EscapedString,
    GeneratorComp,
    GetItem,
    If,
    InlineExpr,
    InputArg,
    LabelConversion,
    NaiveConversion,
    NamedConversion,
    Tuple,
)
from .columns import ColumnRef


class JoinException(Exception):
    pass


class JoinCondition(NamedConversion):
    NAME: str

    def __init__(self):
        super().__init__(self.NAME, GetItem())

    def col(self, column_name: str) -> BaseConversion:

        return self.pipe(ColumnRef(column_name, id_=self.NAME))


class LeftJoinCondition(JoinCondition):
    NAME = "left_row"


class RightJoinCondition(JoinCondition):
    NAME = "right_row"


class _JoinConditions:
    """A helper object to analyze join conditions"""

    LEFT = LeftJoinCondition()
    RIGHT = RightJoinCondition()
    LEFT_NAME = LEFT.name
    RIGHT_NAME = RIGHT.name
    _ANY = {LEFT_NAME, RIGHT_NAME}

    def __init__(self, how="inner"):
        self.inner_loop_conditions = []
        self.left_collection_filters = []
        self.left_row_filters = []
        self.left_row_hashers = []
        self.right_collection_filters = []
        self.right_row_filters = []
        self.right_row_hashers = []
        self.pre_filter = []

        self.inner_join = how == "inner"
        self.left_join = how in {"left", "outer"}
        self.right_join = how in {"right", "outer"}
        self.outer_join = how == "outer"
        if not self.inner_join and not self.left_join and not self.right_join:
            raise AssertionError

    def _add_left_filter(self, filter_conv):
        if self.left_join:
            self.left_row_filters.append(filter_conv)
        else:
            self.left_collection_filters.append(filter_conv)

    def _add_right_filter(self, filter_conv):
        if self.right_join:
            self.right_row_filters.append(filter_conv)
        else:
            self.right_collection_filters.append(filter_conv)

    @classmethod
    def from_condition(cls, condition, **kwargs):
        join_conditions = cls(**kwargs)
        if isinstance(condition, NaiveConversion) and condition.value is True:
            return join_conditions

        if isinstance(condition, Eq):
            join_conditions.consume_eq(condition)
        elif isinstance(condition, And):
            join_conditions.consume_and(condition)
        else:
            join_conditions.consume_other(condition)

        return join_conditions

    @classmethod
    def _get_join_deps(cls, conv) -> t.Set[str]:
        return {
            dep.name
            for dep in conv.get_dependencies(types=NamedConversion)
            if dep.name in cls._ANY
        }

    def consume_eq(self, eq_conversion: Eq):
        if not isinstance(eq_conversion, Eq):
            raise AssertionError

        if len(eq_conversion.args) != 2:
            return self.inner_loop_conditions.append(eq_conversion)

        first_args, second_args = map(self._get_join_deps, eq_conversion.args)
        if len(first_args) > 1 or len(second_args) > 1:
            return self.inner_loop_conditions.append(eq_conversion)

        if len(first_args.union(second_args)) == 2:
            if self.LEFT_NAME in first_args:
                left_idx, right_idx = 0, 1
            else:
                left_idx, right_idx = 1, 0
            self.left_row_hashers.append(eq_conversion.args[left_idx])
            self.right_row_hashers.append(eq_conversion.args[right_idx])
        else:
            all_args = first_args.union(second_args)
            if self.LEFT_NAME in all_args:
                self._add_left_filter(eq_conversion)
            elif self.RIGHT_NAME in all_args:
                self._add_right_filter(eq_conversion)
            else:
                self.pre_filter.append(eq_conversion)

    def consume_other(self, other: BaseConversion):
        deps = self._get_join_deps(other)
        deps_length = len(deps)
        if deps_length > 1:
            self.inner_loop_conditions.append(other)
        elif deps_length == 1:
            if self.LEFT_NAME in deps:
                self._add_left_filter(other)
            else:
                self._add_right_filter(other)
        else:
            self.pre_filter.append(other)

    def consume_and(self, and_conversion: And):
        if not isinstance(and_conversion, And):
            raise AssertionError

        for arg in and_conversion.args:
            if isinstance(arg, Eq):
                self.consume_eq(arg)
            elif isinstance(arg, And):
                self.consume_and(arg)
            else:
                self.consume_other(arg)


class JoinConversion(BaseConversion):
    """Generates conversion which joins left_conversion and right_conversion
    using condition. The result is a generator of joined pairs

    Args:
      left_conversion (BaseConversion): left collection to join
      right_conversion (BaseConversion): right collection to join
      condition (BaseConversion): join condition. If is True, results in cross
        join
      how (str): one of the following: ``"inner"``, ``"left"``, ``"right"``,
        ``"outer"``
    """

    self_content_type = (
        BaseConversion.self_content_type
        | BaseConversion.ContentTypes.NEW_LABEL
    )

    def __init__(
        self,
        left_conversion: BaseConversion,
        right_conversion: BaseConversion,
        condition: BaseConversion,
        how="inner",
    ):
        super().__init__()
        self.left_conversion = self.ensure_conversion(left_conversion)
        self.right_conversion = self.ensure_conversion(right_conversion)
        self.condition = self.ensure_conversion(condition)
        self.how = self.validate_how(how)

    @classmethod
    def validate_how(cls, how: str):
        how = how.lower()
        if how not in ("inner", "left", "right", "outer", "full"):
            raise ValueError(how)
        if how == "full":
            how = "outer"
        return how

    def _gen_code_and_update_ctx(self, code_input, ctx):
        condition = self.condition

        join_conditions = _JoinConditions.from_condition(
            condition, how=self.how
        )
        pre_filter = join_conditions.pre_filter
        inner_loop_conditions = join_conditions.inner_loop_conditions
        if self.how == "right":
            swapped_left_and_right = True
            left_conversion = self.right_conversion
            right_conversion = self.left_conversion
            right_collection_filters = join_conditions.left_collection_filters
            right_row_filters = join_conditions.left_row_filters
            right_row_hashers = join_conditions.left_row_hashers
            left_collection_filters = join_conditions.right_collection_filters
            left_row_filters = join_conditions.right_row_filters
            left_row_hashers = join_conditions.right_row_hashers
        else:
            swapped_left_and_right = False
            left_conversion = self.left_conversion
            right_conversion = self.right_conversion
            left_collection_filters = join_conditions.left_collection_filters
            left_row_filters = join_conditions.left_row_filters
            left_row_hashers = join_conditions.left_row_hashers
            right_collection_filters = join_conditions.right_collection_filters
            right_row_filters = join_conditions.right_row_filters
            right_row_hashers = join_conditions.right_row_hashers

        del join_conditions

        right_collection: BaseConversion = InputArg("right_")
        if right_collection_filters:
            right_collection = right_collection.pipe(
                GeneratorComp(
                    GetItem(),
                    where=(
                        And(
                            right_collection_filters[0],
                            right_collection_filters[1],
                            *right_collection_filters[2:],
                        )
                        if len(right_collection_filters) > 1
                        else right_collection_filters[0]
                    ),
                )
            )
        if self.how == "outer" or not left_row_hashers:
            right_collection = right_collection.as_type(list)
        right_collection = right_collection.add_label("right_collection")

        left_collection: BaseConversion = InputArg("left_")
        if left_collection_filters:
            left_collection = left_collection.pipe(
                GeneratorComp(
                    GetItem(),
                    where=(
                        And(
                            left_collection_filters[0],
                            left_collection_filters[1],
                            *left_collection_filters[2:],
                        )
                        if len(left_collection_filters) > 1
                        else left_collection_filters[0]
                    ),
                )
            )

        inner_loop_condition: t.Optional[ConversionWrapper] = None
        resulting_inner_loop_conditions = list(
            chain(
                left_row_filters,
                right_row_filters,
                inner_loop_conditions,
            )
        )
        if resulting_inner_loop_conditions:
            if len(resulting_inner_loop_conditions) > 1:
                condition_ = And(
                    resulting_inner_loop_conditions[0],
                    resulting_inner_loop_conditions[1],
                    *resulting_inner_loop_conditions[2:],
                )
            else:
                condition_ = resulting_inner_loop_conditions[0]
            inner_loop_condition = ConversionWrapper(
                condition_,
                name_to_code_input={
                    _JoinConditions.LEFT_NAME: "left_item",
                    _JoinConditions.RIGHT_NAME: "right_item",
                },
            )

        if left_row_hashers:
            left_row_conv_to_hash = (
                Tuple(*left_row_hashers)
                if len(left_row_hashers) > 1
                else left_row_hashers[0]
            )
            right_row_conv_to_hash = (
                Tuple(*right_row_hashers)
                if len(right_row_hashers) > 1
                else right_row_hashers[0]
            )

            right_collection_conversion = right_collection.pipe(
                Aggregate(
                    ReduceFuncs.DictArray(
                        right_row_conv_to_hash,
                        GetItem(),
                        default=dict,
                    )
                ),
                label_output="hash_to_items",
            )
            right_items = (
                InlineExpr("left_item").pipe(left_row_conv_to_hash)
            ).pipe(
                If(
                    GetItem().in_(LabelConversion("hash_to_items")),
                    LabelConversion("hash_to_items").item(GetItem()),
                    NaiveConversion(()),
                )
            )
        else:
            right_collection_conversion = right_collection
            right_items = LabelConversion("right_collection")

        if self.how == "inner":
            conv = right_collection_conversion.pipe(
                left_collection.pipe(
                    InlineExpr(
                        "((left_item, right_item)"
                        + " for left_item in {left_items}"
                        + " for right_item in {right_items}"
                        + (" if {condition}" if inner_loop_condition else "")
                        + ")"
                    ).pass_args(
                        left_items=GetItem(),
                        right_items=right_items,
                        condition=inner_loop_condition,
                    ),
                ),
            )
        else:

            if swapped_left_and_right:

                def _left_joiner(left_to_right_ones_gen):
                    none_ = object()
                    for left_item, right_ones in left_to_right_ones_gen:
                        right_item = next(right_ones, none_)
                        if right_item is none_:
                            yield None, left_item
                        else:
                            yield right_item, left_item
                            for right_item in right_ones:
                                yield right_item, left_item

            else:

                def _left_joiner(left_to_right_ones_gen):
                    none_ = object()
                    for left_item, right_ones in left_to_right_ones_gen:
                        right_item = next(right_ones, none_)
                        if right_item is none_:
                            yield left_item, None
                        else:
                            yield left_item, right_item
                            for right_item in right_ones:
                                yield left_item, right_item

            left_joiner = _left_joiner
            conv = right_collection_conversion.pipe(
                left_collection.pipe(
                    InlineExpr(
                        "(left_item, (right_item "
                        "for right_item in {right_items}"
                        + (" if {condition}" if inner_loop_condition else "")
                        + "))"
                        + " for left_item in {left_items}"
                    ).pass_args(
                        left_items=GetItem(),
                        right_items=right_items,
                        condition=inner_loop_condition,
                    ),
                )
            ).pipe(left_joiner)

        if self.how == "outer":

            def _add_right_part(left_join_gen, labels):
                yielded_right_ids = set()
                for left_right in left_join_gen:
                    yielded_right_ids.add(id(left_right[1]))
                    yield left_right

                yield from (
                    (None, right)
                    for right in labels["right_collection"]
                    if id(right) not in yielded_right_ids
                )

            conv = conv.pipe(
                _add_right_part,
                labels=EscapedString("_labels"),
            )
        if pre_filter:
            conv = If(
                (
                    And(
                        pre_filter[0],
                        pre_filter[1],
                        *pre_filter[2:],
                    )
                    if len(pre_filter) > 1
                    else pre_filter[0]
                ),
                conv,
                CallFunc(list),
            )
        code_join = conv.gen_code_and_update_ctx("___TEST___", ctx)
        if "___TEST___" in code_join:
            raise AssertionError(
                "join conversion has a bug, please submit an issue"
            )
        converter_name = self.gen_name("join", ctx, (self, code_input))
        code_args = self.get_args_def_code(as_kwargs=False)
        code = f"""
def {converter_name}(left_, right_{code_args}):
    return {code_join}
        """
        self._code_to_converter(
            converter_name=converter_name, code=code, ctx=ctx
        )
        join_conversion = EscapedString(converter_name).call(
            left_conversion,
            right_conversion,
            *self.get_args_as_func_args(),
        )
        join_conversion.depends_on(self)
        return join_conversion.gen_code_and_update_ctx(code_input, ctx)



"""
The main module exposing public API (via conversion object)
"""
import itertools

from .aggregations import (
    Aggregate,
    BaseReducer,
    GroupBy,
    Reduce,
    ReduceFuncs,
    ReducerDispatcher,
)
from .base import (
    And,
    ApplyFunc,
    BaseConversion,
    CallFunc,
    CodeGenerationOptionsCtx,
    ConversionException,
    ConverterOptionsCtx,
    Dict,
    DictComp,
    DropWhile,
    EscapedString,
    FilterConversion,
    GeneratorComp,
    GetAttr,
    GetItem,
    If,
    InlineExpr,
    InputArg,
    LabelConversion,
    List,
    ListComp,
    NaiveConversion,
    Not,
    OptionalCollectionItem,
    Or,
    Set,
    SetComp,
    SortConversion,
    TakeWhile,
    This,
    Tuple,
    TupleComp,
    ensure_conversion,
)
from .chunks import ChunkBy, ChunkByCondition
from .columns import ColumnRef
from .joins import JoinConversion, _JoinConditions
from .mutations import Mutations


__all__ = ["conversion", "Conversion"]


class Conversion:
    """The object, which exposes public API

    .. code-block:: python

      from convtools import conversion as c

      convert = c.aggregate(
          c.ReduceFuncs.DictSum(
              c.item("name"),
              c.item("value")
          )
      ).gen_converter(debug=True)

      assert convert([
          {"name": "Bob", "value": 10},
          {"name": "Bob", "value": 7},
          {"name": "Ron", "value": 3},
      ]) == {'Bob': 17, 'Ron': 3}

    """

    ConversionException = ConversionException  # pylint: disable=invalid-name
    BaseConversion = BaseConversion  # pylint: disable=invalid-name
    OptionsCtx = ConverterOptionsCtx  # pylint: disable=invalid-name
    CodeGenerationOptionsCtx = (  # pylint: disable=invalid-name
        CodeGenerationOptionsCtx
    )

    ReduceFuncs = ReduceFuncs  # pylint: disable=invalid-name
    #: Shortcut to `Mutations`
    Mut = Mutations  # pylint: disable=invalid-name

    and_ = And
    or_ = Or
    not_ = Not
    if_ = If

    this = This()
    naive = NaiveConversion
    item = GetItem
    attr = GetAttr
    call_func = staticmethod(CallFunc)
    apply_func = staticmethod(ApplyFunc)

    filter = FilterConversion
    sort = SortConversion

    #: Shortcut to `InputArg`
    input_arg = InputArg
    label = LabelConversion
    inline_expr = InlineExpr
    escaped_string = EscapedString

    #: Shortcut to ``List``
    list = List
    tuple = Tuple
    set = Set
    dict = Dict
    optional = OptionalCollectionItem

    generator_comp = GeneratorComp
    iter = GeneratorComp
    list_comp = ListComp
    tuple_comp = TupleComp
    set_comp = SetComp
    dict_comp = DictComp

    group_by = GroupBy
    aggregate = staticmethod(Aggregate)

    join = JoinConversion
    LEFT = _JoinConditions.LEFT
    RIGHT = _JoinConditions.RIGHT

    col = ColumnRef

    take_while = TakeWhile
    drop_while = DropWhile

    chunk_by = ChunkBy
    chunk_by_condition = ChunkByCondition
    CHUNK = ChunkByCondition.CHUNK

    def reduce(self, to_call_with_2_args, *args, **kwargs):
        if (
            isinstance(to_call_with_2_args, type)
            and issubclass(to_call_with_2_args, BaseReducer)
            or isinstance(to_call_with_2_args, ReducerDispatcher)
        ):
            return to_call_with_2_args(*args, **kwargs)
        return Reduce(to_call_with_2_args, *args, **kwargs)

    def __call__(self, obj: object):
        """Shortcut for ``ensure_conversion``"""
        return ensure_conversion(obj)

    def call(self, *args, **kwargs):
        return self.this.call(*args, **kwargs)

    def apply(self, args, kwargs):
        return self.this.apply(args, kwargs)

    def tap(self, *args, **kwargs):
        return self.this.tap(*args, **kwargs)

    def iter_mut(self, *args, **kwargs):
        return self.this.iter_mut(*args, **kwargs)

    def zip(self, *args, **kwargs):
        """Conversion which calls :py:obj:`zip` on conversions.

        Args:
          args: conversions to zip - returns tuples
          kwargs: named conversions to zip - returns dicts

        """
        if args and kwargs:
            raise ValueError("pass either args or kwargs")
        if args:
            return self.call_func(zip, *args)

        return self.call_func(zip, *kwargs.values()).iter(
            {name: self.item(index) for index, name in enumerate(kwargs)}
        )

    def repeat(self, obj, times=None):
        """shortcut to call :py:obj:`itertools.repeat`"""
        args = () if times is None else (times,)
        return self.call_func(itertools.repeat, obj, *args)

    def flatten(self):
        """c.this.flatten() shortcut"""
        return self.this.flatten()

    def min(self, arg1, arg2, *args):
        """c.call_func(min, ...) shortcut"""
        return self.call_func(min, arg1, arg2, *args)

    def max(self, arg1, arg2, *args):
        """c.call_func(max, ...) shortcut"""
        return self.call_func(max, arg1, arg2, *args)

    def breakpoint(self):
        """c.this.breakpoint() shortcut"""
        return self.this.breakpoint()


conversion = Conversion()



"""This module brings aggregations with various reduce functions"""
import typing
from collections import defaultdict
from decimal import Decimal
from math import ceil

from .base import (
    BaseConversion,
    CallFunc,
    CodeGenerationOptionsCtx,
    ConversionException,
    EscapedString,
    FilterConversion,
    GetItem,
    If,
    InlineExpr,
    List,
    NaiveConversion,
    Tuple,
    _None,
    ensure_conversion,
)
from .utils import Code


_none = BaseConversion._none
RBT = typing.TypeVar("RBT", bound="ReduceBlock")


class ReduceBlock:
    """Represents a section of code of a single reducer"""

    var_checksum = "checksum_"
    var_expected_checksum = "expected_checksum_"

    def __init__(
        self,
        reduce_initial: typing.Iterable[str],
        reduce_two: typing.Iterable[str],
        var_row: str,
        var_agg_data_value: str,
        checksum_flag: int,
        unconditional_init: bool,
    ):
        self.reduce_initial = {var_agg_data_value: reduce_initial}
        self.reduce_two = {var_agg_data_value: reduce_two}
        self.var_row = var_row
        self.condition_code = None
        self.unconditional_init = unconditional_init
        self.checksum_flag = checksum_flag

    def consume(self: RBT, reduce_block: RBT) -> RBT:
        if (
            self.condition_code != reduce_block.condition_code
            or self.unconditional_init != reduce_block.unconditional_init
            or self.var_row != reduce_block.var_row
        ):
            raise AssertionError

        for k, v in reduce_block.reduce_initial.items():
            if k in self.reduce_initial:
                raise AssertionError
            self.reduce_initial[k] = v

        for k, v in reduce_block.reduce_two.items():
            if k in self.reduce_two:
                raise AssertionError
            self.reduce_two[k] = v

        self.checksum_flag |= reduce_block.checksum_flag
        return self

    def iter_reduce_lines(self, lines_info_dict) -> typing.Iterable[str]:
        for var_agg_data_value, lines in lines_info_dict.items():
            template_kwargs = {
                "result": var_agg_data_value,
                "row": self.var_row,
            }
            yield from (line % template_kwargs for line in lines)

    def iter_reduce_initial_lines(self) -> typing.Iterable[str]:
        yield from self.iter_reduce_lines(self.reduce_initial)

    def iter_reduce_two_lines(self) -> typing.Iterable[str]:
        yield from self.iter_reduce_lines(self.reduce_two)

    def same_as(self, other):
        return (
            self.condition_code == other.condition_code
            and self.unconditional_init == other.unconditional_init
            and self.var_row == other.var_row
            and len(self.reduce_initial) == len(other.reduce_initial)
            and len(self.reduce_two) == len(other.reduce_two)
            and all(
                l1 == l2
                for l1, l2 in zip(
                    self.reduce_initial.values(), other.reduce_initial.values()
                )
            )
            and all(
                l1 == l2
                for l1, l2 in zip(
                    self.reduce_two.values(), other.reduce_two.values()
                )
            )
        )


class ReduceConditionalBlock(ReduceBlock):
    """Represents a section of code of a single reducer with an incoming
    condition"""

    def __init__(self, *args, **kwargs):
        condition_code = kwargs.pop("condition_code")
        super().__init__(*args, **kwargs)
        self.condition_code = condition_code


class ReduceBlocks(typing.Generic[RBT]):
    """Represents a set of reduce blocks"""

    def __init__(self):
        self.conditional_init_blocks = {}
        self.unconditional_init_blocks = {}
        self.number = 0

    def add_block(self, reduce_block: RBT):
        self.number += 1
        if reduce_block.unconditional_init:
            key_to_block = self.unconditional_init_blocks
        else:
            key_to_block = self.conditional_init_blocks
        key = (
            (
                reduce_block.condition_code
                if isinstance(reduce_block, ReduceConditionalBlock)
                else None
            ),
            reduce_block.var_row,
        )
        if key in key_to_block:
            key_to_block[key].consume(reduce_block)
        else:
            key_to_block[key] = reduce_block

    def iter_blocks(self) -> typing.Iterable[RBT]:
        yield from self.unconditional_init_blocks.values()
        yield from self.conditional_init_blocks.values()

    def gen_group_by_code(
        self, var_row, var_agg_data, var_signature_to_agg_data, code_signature
    ) -> Code:
        code = Code()
        code.add_line(f"for {var_row} in data_:", 1)
        code.add_line(
            f"{var_agg_data} = {var_signature_to_agg_data}[{code_signature}]",
            0,
        )

        for block in self.iter_blocks():
            reduce_initial_lines = list(block.iter_reduce_initial_lines())
            reduce_two_lines = list(block.iter_reduce_two_lines())
            any_var_agg_data_value = next(iter(block.reduce_initial))

            if block.condition_code:
                code.add_line(f"if {block.condition_code}:", 1)

            code.add_line(f"if {any_var_agg_data_value} is _none:", 1)
            for line in reduce_initial_lines:
                code.add_line(line, 0)
            code.incr_indent_level(-1)

            if reduce_two_lines:
                code.add_line("else:", 1)
                for line in reduce_two_lines:
                    code.add_line(line, 0)
                code.incr_indent_level(-1)

            if block.condition_code:
                code.incr_indent_level(-1)
        return code

    def gen_aggregate_code(
        self, var_row, var_checksum, var_expected_checksum
    ) -> Code:
        first_phase_code = Code()
        second_phase_code = Code()

        needs_sum_checking = False

        reduce_blocks = list(self.iter_blocks())
        is_single_block = len(reduce_blocks) == 1
        for block in reduce_blocks:
            reduce_initial_lines = list(block.iter_reduce_initial_lines())
            reduce_two_lines = list(block.iter_reduce_two_lines())
            any_var_agg_data_value = next(iter(block.reduce_initial))

            init_phase_needs_else = not (
                is_single_block and block.unconditional_init
            )

            if block.condition_code:
                first_phase_code.add_line(f"if {block.condition_code}:", 1)

            first_phase_code.add_line(
                f"if {any_var_agg_data_value} is _none:", 1
            )
            for line in reduce_initial_lines:
                first_phase_code.add_line(line, 0)

            if is_single_block:
                if block.unconditional_init:
                    first_phase_code.add_line("break", 0)
                else:
                    first_phase_code.add_line(
                        f"if {any_var_agg_data_value} is not _none:", 1
                    )
                    first_phase_code.add_line("break", -1)

            else:
                if block.unconditional_init:
                    first_phase_code.add_line(
                        f"{var_checksum} |= {block.checksum_flag}", 0
                    )
                else:
                    first_phase_code.add_line(
                        f"if {any_var_agg_data_value} is not _none:", 1
                    )
                    first_phase_code.add_line(
                        f"{var_checksum} |= {block.checksum_flag}", -1
                    )
                needs_sum_checking = True

            first_phase_code.incr_indent_level(-1)

            if init_phase_needs_else and reduce_two_lines:
                first_phase_code.add_line("else:", 1)
                for line in reduce_two_lines:
                    first_phase_code.add_line(line, 0)
                first_phase_code.incr_indent_level(-1)

            if block.condition_code:
                first_phase_code.incr_indent_level(-1)

            if reduce_two_lines:
                if block.condition_code:
                    second_phase_code.add_line(
                        f"if {block.condition_code}:", 1
                    )

                for line in reduce_two_lines:
                    second_phase_code.add_line(line, 0)

                if block.condition_code:
                    second_phase_code.incr_indent_level(-1)

        resulting_code = Code()
        resulting_code.add_line("it_ = iter(data_)", 0)
        resulting_code.add_line(f"for {var_row} in it_:", 1)
        resulting_code.add_code(first_phase_code)
        if needs_sum_checking:
            resulting_code.add_line(
                f"if {var_checksum} == {var_expected_checksum}: break", 0
            )
        resulting_code.incr_indent_level(-1)

        if second_phase_code.has_lines():
            resulting_code.add_line("", 0)
            resulting_code.add_line(f"for {var_row} in it_:", 1)
            resulting_code.add_code(second_phase_code)
            resulting_code.incr_indent_level(-1)
        return resulting_code


GROUPER_TEMPLATE = """
def {converter_name}(data_{code_args}):
    {var_signature_to_agg_data} = defaultdict({var_agg_data_cls})

{code_group_by}

{code_result}
"""
AGGREGATE_TEMPLATE = """
def {converter_name}(data_{code_args}):
    {code_init_agg_vars}
    {var_expected_checksum} = {val_expected_checksum}
    {var_checksum} = 0

{code_aggregate}

{code_result}
"""


RT = typing.TypeVar("RT", bound="BaseReducer")


class BaseReducer(BaseConversion, typing.Generic[RBT]):
    """Base of a reduce operation to be used during the aggregation"""

    method_calls_replace_input_with_self = True

    expressions: typing.Tuple[typing.Any, ...]
    post_conversion: typing.Optional[BaseConversion] = None
    default: typing.Any
    where: typing.Optional[BaseConversion] = None
    unconditional_init: bool = False

    self_content_type = (
        BaseConversion.self_content_type | BaseConversion.ContentTypes.REDUCER
    )

    def gen_reduce_code_block(
        self,
        var_agg_data_value: str,
        var_row: str,
        checksum_flag: int,
        ctx: dict,
    ) -> RBT:
        raise NotImplementedError

    def _gen_code_and_update_ctx(self, code_input, ctx) -> str:
        reducers_run_stage = CodeGenerationOptionsCtx.get_option_value(
            "reducers_run_stage"
        )
        if reducers_run_stage == "collecting_reducer_inputs":
            reducer_inputs_info = ctx["_reducer_inputs_info"][-1]
            if reducer_inputs_info is not False:
                reducer_inputs_info[code_input].append(self)
            agg_data_item = code_input
        elif reducers_run_stage == "rendering_reducer_results":
            overwritten_reducer_inputs = ctx["_overwritten_reducer_inputs"][-1]
            agg_data_item = overwritten_reducer_inputs[(code_input, self)]
        else:
            raise AssertionError(
                "reducers cannot be used outside of aggregations"
            )

        processed_agg_data_item = agg_data_item
        if self.post_conversion:
            processed_agg_data_item = (
                self.post_conversion.gen_code_and_update_ctx(
                    agg_data_item, ctx
                )
            )

        default_value_code = self.default.gen_code_and_update_ctx(None, ctx)
        return EscapedString(
            f"({default_value_code} "
            f"if {agg_data_item} is _none "
            f"else {processed_agg_data_item})"
        ).gen_code_and_update_ctx("", ctx)

    def prepare_default_n_initial(self, default, initial):
        if default is _none:
            try:
                default = self.default
            except AttributeError:
                pass

        if initial is _none and hasattr(self, "initial"):
            initial = self.initial
        if initial is not _none:
            initial = self.ensure_conversion(initial)
            if initial.is_itself_callable_like():
                initial = initial.call_like()

            if default is _none and initial.is_independent():
                default = initial

        if default is _none:
            raise ValueError("default is not provided")

        default = self.ensure_conversion(default)
        if default.is_itself_callable_like():
            default = default.call_like()

        return default, initial


class MultiStatementReducer(BaseReducer):
    """Defines the reduce operation, which is based on multiple python
    statements, to be used during the aggregation"""

    prepare_first: typing.Tuple[str, ...]
    reduce: typing.Tuple[str, ...]

    def __init__(self, *expressions, initial=_none, default=_none, where=None):
        super().__init__()
        self.where = None if where is None else self.ensure_conversion(where)
        self.expressions = tuple(
            self.ensure_conversion(expr)
            for expr in (
                self.expressions
                if hasattr(self, "expressions") and not expressions
                else expressions
            )
        )
        self.default, self.initial = self.prepare_default_n_initial(
            default, initial
        )

    def _format_statements(
        self,
        statements,
        args,
        var_row,
        ctx,
        prev_result="%(result)s",
    ):
        code_args = tuple(
            arg.gen_code_and_update_ctx(var_row, ctx) for arg in args
        )
        return [
            statement.format(*code_args, prev_result=prev_result)
            for statement in statements
        ]

    def gen_reduce_code_block(
        self,
        var_agg_data_value: str,
        var_row: str,
        checksum_flag: int,
        ctx: dict,
    ) -> RBT:
        if self.initial is _none:
            reduce_initial = self._format_statements(
                self.prepare_first,
                self.expressions,
                var_row,
                ctx,
            )
        else:
            reduce_initial = self._format_statements(
                self.reduce,
                self.expressions,
                var_row,
                ctx,
                prev_result=self.initial.gen_code_and_update_ctx(var_row, ctx),
            )

        reduce_two = self._format_statements(
            self.reduce,
            self.expressions,
            var_row,
            ctx,
        )

        block_cls = (
            ReduceBlock if self.where is None else ReduceConditionalBlock
        )
        kwargs = {
            "reduce_initial": reduce_initial,
            "reduce_two": reduce_two,
            "var_row": var_row,
            "var_agg_data_value": var_agg_data_value,
            "checksum_flag": checksum_flag,
            "unconditional_init": self.unconditional_init,
        }
        if self.where is not None:
            kwargs["condition_code"] = self.where.gen_code_and_update_ctx(
                var_row, ctx
            )

        return block_cls(**kwargs)


class Reduce(BaseReducer):
    """Defines the reduce operation, which is based on a callable / expression
    to be used during the aggregation"""

    initial: typing.Any

    def __init__(
        self,
        to_call_with_2_args: typing.Union[typing.Callable, InlineExpr],
        *expressions: typing.Tuple[typing.Any, ...],
        initial: typing.Union[_None, typing.Callable, InlineExpr, typing.Any],
        default: typing.Union[_None, typing.Callable, typing.Any] = _none,
        unconditional_init: bool = False,
        where=None,
    ):
        """
        Args:
          to_call_with_2_args: defines the reduce function/expression
          expressions: args to be passed to `to_call_with_2_args` after the
            aggregation value
          initial: defines the very first item to be passed to
            `to_call_with_2_args` item. If callable, then the result of a call
            is used. If a conversion is passed, it is resolved on the first
            row met.
          default: defines the value to be returned when there was nothing to
            reduce in a group (e.g. the current reduce operation has filtered
            out some rows, while an adjacent reduce operation has got
            something to reduce, forming a group). If callable, then the result
            of a call is used.  When default is not passed, initial is used if
            it doesn't depend on input data.
          unconditional_init: tells whether the first call initializes the
            aggregation value OR there is a condition for that
        """
        super().__init__()
        self.where = None if where is None else self.ensure_conversion(where)
        self.to_call_with_2_args = self.ensure_conversion(to_call_with_2_args)
        self.expressions = tuple(
            self.ensure_conversion(expr) for expr in expressions
        )
        self.default, self.initial = self.prepare_default_n_initial(
            default, initial
        )
        self.unconditional_init = unconditional_init

    def gen_reduce_code_block(
        self,
        var_agg_data_value: str,
        var_row: str,
        checksum_flag: int,
        ctx: dict,
    ) -> RBT:
        _ = self.to_call_with_2_args.call_like(
            self.initial,
            *self.expressions,
        ).gen_code_and_update_ctx(var_row, ctx)
        reduce_initial = [f"%(result)s = {_}"]
        _ = self.to_call_with_2_args.call_like(
            EscapedString("%(result)s"),
            *self.expressions,
        ).gen_code_and_update_ctx(var_row, ctx)
        reduce_two = [f"%(result)s = {_}"]

        block_cls = (
            ReduceBlock if self.where is None else ReduceConditionalBlock
        )
        kwargs = {
            "reduce_initial": reduce_initial,
            "reduce_two": reduce_two,
            "var_row": var_row,
            "var_agg_data_value": var_agg_data_value,
            "checksum_flag": checksum_flag,
            "unconditional_init": self.unconditional_init,
        }
        if self.where is not None:
            kwargs["condition_code"] = self.where.gen_code_and_update_ctx(
                var_row, ctx
            )

        return block_cls(**kwargs)


class GroupBy(BaseConversion, typing.Generic[RBT]):
    """Generates the function which aggregates the data, grouping by
    conversions, specified in `__init__` method and returns list of items in a
    format defined by the parameter passed to ``aggregate`` method.

    If no group keys are passed, then it returns just a single value, defined
    by the parameter passed to ``aggregate`` method.

    Current optimizations:
     * piping like ``c.group_by(...).aggregate().pipe(...)`` won't run
       the aggregation twice, this is handled as 2 statements
     * using the same reduce clause twice (e.g. one used as an argument
       for some function calls) won't result in calculating this reduce twice
    """

    self_content_type = (
        BaseConversion.self_content_type
        | BaseConversion.ContentTypes.AGGREGATION
    )

    def __init__(self, *by: typing.Tuple[BaseConversion, ...]):
        """Takes any number of conversions to group by

        Args:
          by (tuple): each item is to be wrapped with
            :py:obj:`ensure_conversion`.  Each is to resolve to a hashable
            object to allow using such tuples as keys. If nothing is passed,
            aggregate the input into a single object.
        """
        super().__init__()
        self.by = [self.ensure_conversion(by_) for by_ in by]
        self.agg_result: typing.Optional[BaseConversion] = None
        self.aggregate_mode = len(self.by) == 0
        self.filter_conversion = None
        self.filter_cast = None

    def aggregate(
        self, reducer: typing.Union[dict, list, set, tuple, BaseConversion]
    ) -> "GroupBy":
        """Takes the conversion which defines the desired output of
        aggregation"""
        self_clone = self.clone()
        self_clone.agg_result = self_clone.ensure_conversion(reducer)
        self_clone.contents = self_clone.contents & ~self.ContentTypes.REDUCER
        if isinstance(self_clone.agg_result, NaiveConversion):
            raise AssertionError("unexpected reducer type", type(reducer))
        return self_clone

    def filter(
        self, condition_conv, cast=BaseConversion._none
    ) -> "BaseConversion":
        if cast is self._none:
            cast = list
        if self.aggregate_mode or self.filter_conversion is not None:
            return super().filter(condition_conv, cast=cast)
        self_clone = self.clone()
        self_clone.filter_conversion = self_clone.ensure_conversion(
            condition_conv
        )
        self_clone.filter_cast = cast
        return self_clone

    def _gen_agg_data_container(self, container_name, number_of_reducers, ctx):
        attrs = [f"v{i}" for i in range(number_of_reducers)]

        code = Code()
        code.add_line(f"class {container_name}:", 1)
        _ = ",".join(f"'{attr}'" for attr in attrs)
        code.add_line(f"__slots__ = [{_}]", 0)
        code.add_line("def __init__(self, _none=__none__):", 1)
        if attrs:
            for attr in attrs:
                code.add_line(f"self.{attr} = _none", 0)
        else:
            code.add_line("pass", 0)
        return self._code_to_converter(container_name, code.to_string(0), ctx)

    def _gen_code_and_update_ctx(self, code_input, ctx) -> str:
        if self.agg_result is None:
            raise AssertionError("aggregate hasn't been called")
        aggregate_mode = len(self.by) == 0
        suffix = self.gen_name("_", ctx, self)
        var_row = f"row{suffix}"
        var_signature = f"signature{suffix}"
        var_signature_to_agg_data = f"signature_to_agg_data{suffix}"
        var_agg_data = f"agg_data{suffix}"
        var_agg_data_cls = f"AggData{suffix}"

        signature_code_items = [
            by_.gen_code_and_update_ctx(var_row, ctx) for by_ in self.by
        ]
        replacements = {}

        # remembering code replacements for group by keys
        if len(signature_code_items) == 1:
            code_signature = signature_code_items[0]
            replacements[code_signature] = var_signature
        else:
            code_signature = f"({','.join(signature_code_items)},)"
            replacements.update(
                {
                    code: f"{var_signature}[{index}]"
                    for index, (code, by_) in enumerate(
                        zip(signature_code_items, self.by)
                    )
                }
            )

        # collecting inputs of reducers and reducers themselves
        with CodeGenerationOptionsCtx() as options:
            options.reducers_run_stage = "collecting_reducer_inputs"
            key = "_reducer_inputs_info"
            if key not in ctx:
                ctx[key] = []
            ctx[key].append(defaultdict(list))
            self.agg_result.gen_code_and_update_ctx(var_row, ctx)
            reducer_inputs_info = ctx[key].pop()

        def gen_agg_data_value(value_index):
            if aggregate_mode:
                return EscapedString(
                    f"{var_agg_data}_v{value_index}"
                ).gen_code_and_update_ctx("", ctx)
            else:
                return (
                    EscapedString(var_agg_data)
                    .attr(f"v{value_index}")
                    .gen_code_and_update_ctx("", ctx)
                )

        expected_checksum = 0
        var_agg_data_values = []

        blocks: typing.List[RBT] = []
        overwritten_reducer_inputs = {}
        # reusing same reducers, remembering code replacements for reducers
        for reducer_code_input, reducer in (
            (reducer_code_input, reducer)
            for reducer_code_input, reducers in reducer_inputs_info.items()
            for reducer in reducers
        ):
            reduce_block_index = len(blocks)
            checksum_flag = (
                1 << reduce_block_index if self.aggregate_mode else 0
            )
            var_agg_data_value = gen_agg_data_value(reduce_block_index)
            reduce_block = reducer.gen_reduce_code_block(
                var_agg_data_value,
                reducer_code_input,
                checksum_flag,
                ctx,
            )

            such_block_exists = False
            for index, block in enumerate(blocks):
                if reduce_block.same_as(block):
                    reduce_block_index = index
                    such_block_exists = True
                    break

            if not such_block_exists:
                blocks.append(reduce_block)
                var_agg_data_values.append(var_agg_data_value)
                expected_checksum |= checksum_flag

            overwritten_reducer_inputs[
                (reducer_code_input, reducer)
            ] = gen_agg_data_value(reduce_block_index)

        reduce_blocks: ReduceBlocks = ReduceBlocks()
        for block in blocks:
            reduce_blocks.add_block(block)

        # populates reducers with their code inputs
        with CodeGenerationOptionsCtx() as options:
            options.reducers_run_stage = "rendering_reducer_results"
            key = "_overwritten_reducer_inputs"
            if key not in ctx:
                ctx[key] = []
            ctx[key].append(overwritten_reducer_inputs)
            code_agg_result = self.agg_result.gen_code_and_update_ctx(
                var_row, ctx
            )
            ctx[key].pop()

        ctx.update({"defaultdict": defaultdict})

        if aggregate_mode:
            _ = " = ".join(var_agg_data_values)
            code_init_agg_vars = f"{_} = _none"
        else:
            ctx[var_agg_data_cls] = self._gen_agg_data_container(
                var_agg_data_cls, reduce_blocks.number, ctx
            )

        # used by SortedArrayReducer
        ctx["ListSortedOnceWrapper"] = ListSortedOnceWrapper

        for code_from, code_to in replacements.items():
            code_agg_result = self.replace_word(
                code_agg_result, code_from, code_to
            )

        if self.count_words(code_agg_result, var_row):
            raise ConversionException(
                "neither group by key nor a field used in a reducer",
                code_agg_result,
            )
        if self.aggregate_mode:
            code_agg_result = f"    return {code_agg_result}"
        else:
            if self.filter_conversion is None:
                code_agg_result = (
                    f"    return [{code_agg_result} "
                    f"for {var_signature}, {var_agg_data} "
                    f"in {var_signature_to_agg_data}.items()]"
                )
            else:
                code_filtered_result = FilterConversion(
                    self.filter_conversion,
                    self.filter_cast,
                ).gen_code_and_update_ctx("result_", ctx)
                code_agg_result = (
                    f"    result_ = ({code_agg_result} "
                    f"for {var_signature}, {var_agg_data} "
                    f"in {var_signature_to_agg_data}.items())\n"
                    f"    filtered_result_ = {code_filtered_result}\n"
                ) + (
                    "    yield from filtered_result_"
                    if self.filter_cast is None
                    else "    return filtered_result_"
                )

        agg_template_kwargs = dict(
            code_args=self.get_args_def_code(as_kwargs=False),
            code_result=code_agg_result,
            var_row=var_row,
            expected_checksum=expected_checksum,
        )

        if self.aggregate_mode:
            converter_name = f"aggregate{suffix}"
            grouper_code = AGGREGATE_TEMPLATE.format(
                converter_name=converter_name,
                code_init_agg_vars=code_init_agg_vars,
                var_expected_checksum=ReduceBlock.var_expected_checksum,
                val_expected_checksum=expected_checksum,
                var_checksum=ReduceBlock.var_checksum,
                code_aggregate=reduce_blocks.gen_aggregate_code(
                    var_row=var_row,
                    var_checksum=ReduceBlock.var_checksum,
                    var_expected_checksum=expected_checksum,
                ).to_string(
                    base_indent_level=1,
                ),
                **agg_template_kwargs,
            )
        else:
            converter_name = f"group_by{suffix}"
            grouper_code = GROUPER_TEMPLATE.format(
                converter_name=converter_name,
                var_signature_to_agg_data=var_signature_to_agg_data,
                var_agg_data_cls=var_agg_data_cls,
                var_agg_data=var_agg_data,
                code_signature=code_signature,
                code_group_by=reduce_blocks.gen_group_by_code(
                    var_row=var_row,
                    var_agg_data=var_agg_data,
                    var_signature_to_agg_data=var_signature_to_agg_data,
                    code_signature=code_signature,
                ).to_string(base_indent_level=1),
                **agg_template_kwargs,
            )

        group_data_func = self._code_to_converter(
            converter_name=converter_name,
            code=grouper_code,
            ctx=ctx,
        )
        return CallFunc(
            group_data_func, GetItem(), *self.get_args_as_func_args()
        ).gen_code_and_update_ctx(code_input, ctx)


def Aggregate(  # pylint:disable=invalid-name
    *args, **kwargs
) -> BaseConversion:
    """Shortcut for ``GroupBy().aggregate(*args, **kwargs)``"""
    return GroupBy().aggregate(*args, **kwargs)


class ReducerDispatcher:
    pass


class SumReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = ({0} or 0)",)
    reduce = ("%(result)s = {prev_result} + ({0} or 0)",)
    default = NaiveConversion(0)
    unconditional_init = True


class FastSumReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = ({0} or 0)",)
    reduce = ("%(result)s = {prev_result} + {0}",)
    default = NaiveConversion(0)
    unconditional_init = True


class SumReducerDispatcher(ReducerDispatcher):
    def __call__(self, expression, *args, **kwargs):
        expression = ensure_conversion(expression)
        if expression.output_hints & expression.OutputHints.NOT_NONE:
            return FastSumReducer(expression, *args, **kwargs)
        return SumReducer(expression, *args, **kwargs)


class SumOrNoneReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = {0}",)
    reduce = (
        "if {0} is None:",
        "    %(result)s = None",
        "elif %(result)s is not None:",
        "    %(result)s = {prev_result} + {0}",
    )
    default = NaiveConversion(None)
    unconditional_init = True


class MaxReducer(MultiStatementReducer):
    prepare_first = (
        "if {0} is not None:",
        "    %(result)s = {0}",
    )
    reduce = (
        "if {0} is not None and {0} > %(result)s:",
        "    %(result)s = {0}",
    )
    default = NaiveConversion(None)


class MinReducer(MultiStatementReducer):
    prepare_first = (
        "if {0} is not None:",
        "    %(result)s = {0}",
    )
    reduce = (
        "if {0} is not None and {0} < %(result)s:",
        "    %(result)s = {0}",
    )
    default = NaiveConversion(None)


class CountReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = 1",)
    reduce = ("%(result)s = {prev_result} + 1",)
    default = NaiveConversion(0)
    unconditional_init = True


class CountDistinctReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = {{ {0} }}",)
    reduce = ("%(result)s.add({0})",)
    default = NaiveConversion(0)
    unconditional_init = True
    post_conversion = CallFunc(len, GetItem())


class FirstReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = {0}",)
    reduce = ()
    default = NaiveConversion(None)
    unconditional_init = True


class LastReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = {0}",)
    reduce = ("%(result)s = {0}",)
    default = NaiveConversion(None)
    unconditional_init = True


class MaxRowReducer(MultiStatementReducer):
    """Reducer which finds an item with max value of the expression and returns
    this item"""

    prepare_first = (
        "if {0} is not None:",
        "    %(result)s = ({0}, %(row)s)",
    )
    reduce = (
        "if {0} is not None and %(result)s[0] < {0}:",
        "    %(result)s = ({0}, %(row)s)",
    )
    default = NaiveConversion(None)
    post_conversion = GetItem(1)


class MinRowReducer(MaxRowReducer):
    reduce = (
        "if {0} is not None and %(result)s[0] > {0}:",
        "    %(result)s = ({0}, %(row)s)",
    )


class ArrayReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = [{0}]",)
    reduce = ("%(result)s.append({0})",)
    default = NaiveConversion(None)
    unconditional_init = True


class ListSortedOnceWrapper:
    """Wraps a list, exposes append method only. Once the list is filled up, it
    is sorted (only once) in-place and is returned when get method called."""

    __slots__ = ["list_", "append", "sorted"]

    def __init__(self, list_: list):
        self.list_ = list_
        self.append = self.list_.append
        self.sorted = False

    def get(self) -> list:
        if not self.sorted:
            self.list_.sort()
            self.sorted = True
            del self.append
        return self.list_


class SortedArrayReducer(MultiStatementReducer):
    # preserve this extra line to keep this different from ArrayReducer
    prepare_first = ("%(result)s = ListSortedOnceWrapper([{0}])", "")
    reduce = ("%(result)s.append({0})",)
    default = NaiveConversion(None)
    unconditional_init = True

    @property
    def post_conversion(self):
        return GetItem().call_method("get")


class ArrayDistinctReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = {{ {0}: None }}",)
    reduce = ("%(result)s[{0}] = None",)
    post_conversion = InlineExpr("list({0})").pass_args(GetItem())
    default = NaiveConversion(None)
    unconditional_init = True


class BaseDictReducer(MultiStatementReducer):
    """This reducer accepts 2 expressions:

    - the first one is used to calculate keys of the resulting dict
    - the second one is used to calculate values to be reduced and put as the
      final value, under the certain key.

    Effectively dict reducers allow for double grouping: the first one on the
    top level, the second one on DictReducer level.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # for backward compatibility with versions <= 0.8
        if len(self.expressions) == 1:
            expr = self.expressions[0]
            if isinstance(expr, (Tuple, List)) and len(expr.items) == 2:
                self.expressions = tuple(expr.items)
        if len(self.expressions) != 2:
            raise ValueError("invalid dict reducer input: two values expected")


class DictReducer(BaseDictReducer):
    prepare_first = ("%(result)s = {{ {0}: {1} }}",)
    reduce = ("%(result)s[{0}] = {1}",)
    default = NaiveConversion(None)
    unconditional_init = True


class DictArrayReducer(BaseDictReducer):
    prepare_first = (
        "%(result)s = _d = defaultdict(list)",
        "_d[{0}].append({1})",
    )
    reduce = ("%(result)s[{0}].append({1})",)
    post_conversion = InlineExpr("dict({})").pass_args(GetItem())
    default = NaiveConversion(None)
    unconditional_init = True


class DictArrayDistinctReducer(BaseDictReducer):
    """Dict reducer where dict values are lists of distinct values"""

    prepare_first = (
        "%(result)s = _d = defaultdict(dict)",
        "_d[{0}][{1}] = None",
    )
    reduce = ("%(result)s[{0}][{1}] = None",)
    post_conversion = InlineExpr(
        "{{k_: list(v_) for k_, v_ in {}.items()}}"
    ).pass_args(GetItem())
    default = NaiveConversion(None)
    unconditional_init = True


class DictSumReducer(BaseDictReducer):
    prepare_first = (
        "%(result)s = _d = defaultdict(int)",
        "_d[{0}] = {1} or 0",
    )
    reduce = ("%(result)s[{0}] = {prev_result}[{0}] + ({1} or 0)",)
    post_conversion = InlineExpr("dict({})").pass_args(GetItem())
    default = NaiveConversion(None)
    unconditional_init = True


class DictSumOrNoneReducer(BaseDictReducer):
    """Dict reducer where dict values are either numbers or None if there's
    been at least one None value within the group"""

    prepare_first = ("%(result)s = _d = defaultdict(int)", "_d[{0}] = {1}")
    reduce = (
        "if {1} is None:",
        "    %(result)s[{0}] = None",
        "elif %(result)s[{0}] is not None:",
        "    %(result)s[{0}] = {prev_result}[{0}] + {1}",
    )
    post_conversion = InlineExpr("dict({})").pass_args(GetItem())
    default = NaiveConversion(None)
    unconditional_init = True


class DictMaxReducer(BaseDictReducer):
    """DictMax reducer which takes first positional item as keys and
    accumulates max value of second positional item"""

    prepare_first = (
        "if {1} is not None:",
        "    %(result)s = {{ {0}: {1} }}",
    )
    reduce = (
        "if {1} is not None and "
        "({0} not in %(result)s or {1} > %(result)s[{0}]):",
        "    %(result)s[{0}] = {1}",
    )
    default = NaiveConversion(None)


class DictMinReducer(DictMaxReducer):
    reduce = (
        "if {1} is not None and "
        "({0} not in %(result)s or {1} < %(result)s[{0}]):",
        "    %(result)s[{0}] = {1}",
    )


class DictCountReducer(BaseDictReducer):
    prepare_first = ("%(result)s = {{ {0}: 1 }}",)
    reduce = (
        "if {0} not in %(result)s:",
        "    %(result)s[{0}] = 1",
        "else:",
        "    %(result)s[{0}] = {prev_result}[{0}] + 1",
    )
    default = NaiveConversion(None)
    unconditional_init = True


class DictCountDistinctReducer(BaseDictReducer):
    """Dict reducer where dict values are numbers of distinct values per
    group"""

    prepare_first = ("%(result)s = {{ {0}: {{ {1} }} }}",)
    reduce = (
        "if {0} not in %(result)s:",
        "    %(result)s[{0}] = {{ {1} }}",
        "else:",
        "    %(result)s[{0}].add({1})",
    )
    post_conversion = InlineExpr(
        "{{ k_: len(v_) for k_, v_ in {}.items() }}"
    ).pass_args(GetItem())
    default = NaiveConversion(None)
    unconditional_init = True


class DictFirstReducer(BaseDictReducer):
    prepare_first = ("%(result)s = {{ {0}: {1} }}",)
    reduce = ("if {0} not in %(result)s:", "    %(result)s[{0}] = {1}")
    default = NaiveConversion(None)
    unconditional_init = True


class DictLastReducer(BaseDictReducer):
    prepare_first = ("%(result)s = {{ {0}: {1} }}",)
    reduce = ("%(result)s[{0}] = {1}",)
    default = NaiveConversion(None)
    unconditional_init = True


class AverageReducerDispatcher(ReducerDispatcher):
    """Dispatcher which chooses between weighted and simple averages"""

    sum_reducer_dispatcher = SumReducerDispatcher()

    def __call__(
        self, value, weight=1, default=None, where=None
    ) -> "BaseConversion":
        """
        Calculates the arithmetic mean or weighted mean.
        """
        if isinstance(weight, (int, float, Decimal)) and weight == 1:
            return If(
                CountReducer(where=where),
                (
                    self.sum_reducer_dispatcher(value, where=where)
                    / CountReducer(where=where)
                ),
                default,
            )
        return If(
            self.sum_reducer_dispatcher(weight, where=where),
            (
                self.sum_reducer_dispatcher(value * weight, where=where)
                / self.sum_reducer_dispatcher(weight, where=where)
            ),
            default,
        )


class TopReducer(DictCountReducer):
    """
    Returns a list of the most frequent values.
    The resulting list is sorted in descending order of values frequency.
    """

    def __init__(self, k: int, key_conv, *args, **kwargs):
        super().__init__(key_conv, 1, *args, **kwargs)
        if not isinstance(k, int):
            raise TypeError("K must be an integer.")

        if k < 1:
            raise ValueError("K must be a positive integer greater than 0.")

        self.k = k

    @property
    def post_conversion(self):
        return InlineExpr(
            "[k for k,v in sorted((v,k) for k,v in {data}.items())[:-{k}:-1]]"
        ).pass_args(data=GetItem(), k=self.k + 1)


class ModeReducer(DictCountReducer):
    def __init__(self, conv, *args, **kwargs):
        super().__init__(conv, conv, *args, **kwargs)

    post_conversion = InlineExpr(
        "sorted(((v,k) for k,v in {data}.items()))[-1][1]"
    ).pass_args(data=GetItem())


class PercentileReducer(SortedArrayReducer):
    """Calculates percentile (floats from 0 to 100 inclusive)

    >>> c.ReduceFuncs.Percentile(95, c.item("amount"))
    >>> c.ReduceFuncs.Percentile(95, c.item("amount"), interpolation="lower")

    interpolation options:
      * "linear"
      * "lower"
      * "higher"
      * "midpoint"
      * "nearest"
    """

    interpolation_to_method: "typing.Dict[str, typing.Callable]" = {}

    def __init__(
        self, percentile: float, conv, *args, interpolation="linear", **kwargs
    ):
        """

        Args:
         * interpolation: one of
           #. "linear"
           #. "lower"
           #. "higher"
           #. "midpoint"
           #. "nearest"
        """
        super().__init__(conv, *args, **kwargs)
        if not 0 <= percentile <= 100:
            raise ValueError(
                "percentile must be a float between 0 and 100 inclusive"
            )

        self.percentile = percentile
        try:
            self.method = self.interpolation_to_method[interpolation]
        except KeyError as e:
            raise ValueError("unsupported interpolation type") from e

    @staticmethod
    def percentile_linear(data, quantile):
        max_index = len(data) - 1
        index = max_index * quantile
        left_index = int(index)
        left_value = data[left_index]
        if left_index == max_index:
            return left_value

        return left_value + (data[left_index + 1] - left_value) * (
            index - left_index
        )

    @staticmethod
    def percentile_lower(data, quantile):
        return data[int((len(data) - 1) * quantile)]

    @staticmethod
    def percentile_higher(data, quantile):
        return data[ceil((len(data) - 1) * quantile)]

    @staticmethod
    def percentile_midpoint(data, quantile):
        index = (len(data) - 1) * quantile
        left_index = int(index)
        if left_index == index:
            return data[left_index]

        left_value = data[left_index]
        return left_value + (data[left_index + 1] - left_value) * 0.5

    @staticmethod
    def percentile_nearest(data, quantile):
        index = (len(data) - 1) * quantile
        left_index = int(index)
        if index - left_index > 0.5:
            return data[left_index + 1]
        else:
            return data[left_index]

    @property
    def post_conversion(self):
        return CallFunc(
            self.method, super().post_conversion, self.percentile * 0.01
        )


PercentileReducer.interpolation_to_method = {
    "linear": PercentileReducer.percentile_linear,
    "lower": PercentileReducer.percentile_lower,
    "higher": PercentileReducer.percentile_higher,
    "midpoint": PercentileReducer.percentile_midpoint,
    "nearest": PercentileReducer.percentile_nearest,
}


def MedianReducer(  # pylint:disable=invalid-name
    conv, *args, **kwargs
) -> BaseConversion:
    return PercentileReducer(50, conv, *args, **kwargs)


class ReduceFuncs:
    """Exposes the list of reduce functions"""

    # pylint: disable=invalid-name

    #: Calculates the sum, skips false values
    Sum = SumReducerDispatcher()
    #: Calculates the sum, any ``None`` makes the total sum ``None``
    SumOrNone = SumOrNoneReducer

    #: Finds max value, skips ``None``
    Max = MaxReducer
    #: Finds a row with max value, skips ``None``
    MaxRow = MaxRowReducer

    #: Finds min value, skips ``None``
    Min = MinReducer
    #: Finds a row with min value, skips ``None``
    MinRow = MinRowReducer

    #: Counts objects
    Count = CountReducer
    #: Counts distinct values
    CountDistinct = CountDistinctReducer

    #: Stores the first value per group
    First = FirstReducer
    #: Stores the last value per group
    Last = LastReducer

    #: Calculates the arithmetic mean or weighted mean.
    Average = AverageReducerDispatcher()
    #: Calculates the median value.
    Median = MedianReducer
    #: Calculates percentile: floats in [0, 100]
    Percentile = PercentileReducer
    #: Calculates the most common value.
    #: In case of multiple values, returns the last of them.
    Mode = ModeReducer
    #: Returns a list of the most frequent values.
    #: The resulting list is sorted in descending order of values frequency.
    TopK = TopReducer

    #: Aggregates values into array
    Array = ArrayReducer
    #: Aggregates distinct values into array, preserves order
    ArrayDistinct = ArrayDistinctReducer

    #: Aggregates values into dict; dict values are last values per group
    Dict = DictReducer
    #: Aggregates values into dict; dict values are lists of group values
    DictArray = DictArrayReducer
    #: Aggregates values into dict; dict values are lists of unique group
    #: values preserves order
    DictArrayDistinct = DictArrayDistinctReducer
    #: Aggregates values into dict; dict values are sums of group values,
    #: skipping ``None``
    DictSum = DictSumReducer
    #: Aggregates values into dict; dict values are sums of group values,
    #: any ``None`` makes the total sum ``None``
    DictSumOrNone = DictSumOrNoneReducer
    #: Aggregates values into dict; dict values are max group values
    DictMax = DictMaxReducer
    #: Aggregates values into dict; dict values are min group values
    DictMin = DictMinReducer
    #: Aggregates values into dict; dict values are numbers of values in groups
    DictCount = DictCountReducer
    #: Aggregates values into dict; dict values are numbers of unique values
    #: in groups
    DictCountDistinct = DictCountDistinctReducer
    #: Aggregates values into dict; dict values are first values per group
    DictFirst = DictFirstReducer
    #: Aggregates values into dict; dict values are last values per group
    DictLast = DictLastReducer



"""Helpers live here: like:
 - recently used cache
 - options ctx manager
"""
import sys
import threading
import typing


if sys.version_info[:2] == (3, 6):
    GenericMeta = typing.GenericMeta
else:

    class GenericMeta(type):
        pass


class BaseCtxMeta(GenericMeta):
    def __init__(cls, name, bases, kwargs):
        super().__init__(name, bases, kwargs)
        cls._ctx = threading.local()


class BaseOptionsMeta(type):
    def __init__(cls, name, bases, kwargs):
        super().__init__(name, bases, kwargs)
        cls._option_attrs = {
            k: v
            for k, v in kwargs.items()
            if not k.startswith("_") and k not in {"clone", "to_defaults"}
        }


class BaseOptions(object, metaclass=BaseOptionsMeta):
    """Container object, which carries current options"""

    def clone(self):
        clone = self.__class__()
        for option_attr in self._option_attrs.keys():
            setattr(clone, option_attr, getattr(self, option_attr))
        return clone

    def to_defaults(self, option_name=None):
        if option_name:
            setattr(self, option_name, self._option_attrs[option_name])
        else:
            for option_attr, value in self._option_attrs.items():
                setattr(self, option_attr, value)


OT = typing.TypeVar("OT", bound=BaseOptions)


class BaseCtx(
    typing.Generic[OT], metaclass=BaseCtxMeta
):  # pylint:disable=invalid-metaclass
    """Context manager to manage option objects"""

    options_cls: typing.Type[OT]
    _ctx: threading.local

    def __enter__(self) -> OT:
        self._ctx.prev_options = getattr(self._ctx, "options", None)
        if self._ctx.prev_options:
            self._ctx.options = self._ctx.prev_options.clone()
        else:
            self._ctx.options = self.options_cls()
        return self._ctx.options

    def __exit__(self, exc_type, exc_value, tb):
        self._ctx.options = self._ctx.prev_options
        self._ctx.prev_options = None

    @classmethod
    def get_option_value(cls, option_name):
        options = getattr(cls._ctx, "options", None)
        if not options:
            options = cls.options_cls
        return getattr(options, option_name)


class Code:
    """A building block for generating code, which is composed of multiple
    statements."""

    def __init__(self):
        self.lines_info = []
        self.indent_level = 0

    def add_line(self, line: str, next_line_indent_incr: int):
        self.lines_info.append((self.indent_level, line))
        self.indent_level += next_line_indent_incr
        if self.indent_level < 0:
            raise AssertionError("negative indentation level")

    def add_code(self, code: "Code"):
        for indent_level, line in code.lines_info:
            self.lines_info.append((indent_level + self.indent_level, line))

    def incr_indent_level(self, incr: int):
        self.indent_level += incr

    def has_lines(self):
        return bool(self.lines_info)

    def to_string(self, base_indent_level: int, single_indent: str = "    "):
        return "\n".join(
            f"{single_indent * (base_indent_level + indent_level)}{line}"
            for indent_level, line in self.lines_info
        )



"""
Implements streaming operations on table-like data and csv files.

Conversions are defined in realtime based on table headers and called methods:
 - update
 - take
 - drop
 - join

"""
import csv
import typing as t  # pylint: disable=unused-import
from itertools import chain, zip_longest

from ..base import (
    And,
    BaseConversion,
    ConverterOptionsCtx,
    GeneratorComp,
    GetItem,
    If,
    InputArg,
    NaiveConversion,
    ensure_conversion,
)
from ..columns import ColumnChanges, ColumnRef, MetaColumns
from ..joins import JoinConversion, LeftJoinCondition, RightJoinCondition


class CloseFileIterator:
    """
    Utility to close the corresponding file once the iterator is exhausted
    """

    def __init__(self, file_to_close):
        self.file_to_close = file_to_close

    def __iter__(self):
        return self

    def __next__(self):
        self.file_to_close.close()
        self.file_to_close = None
        raise StopIteration

    def __del__(self):
        if self.file_to_close:
            self.file_to_close.close()
            self.file_to_close = None


class CustomCsvDialect(csv.Dialect):
    """A helper to define custom csv dialects without defining classes"""

    def __init__(
        self,
        delimiter=csv.excel.delimiter,
        quotechar=csv.excel.quotechar,
        escapechar=csv.excel.escapechar,
        doublequote=csv.excel.doublequote,
        skipinitialspace=csv.excel.skipinitialspace,
        lineterminator=csv.excel.lineterminator,
        quoting=csv.excel.quoting,
    ):
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.escapechar = escapechar
        self.doublequote = doublequote
        self.skipinitialspace = skipinitialspace
        self.lineterminator = lineterminator
        self.quoting = quoting
        super().__init__()


class Table:
    """Table conversion exposes streaming operations on table-like data and csv
    files

    >>> Table.from_csv("input.csv", header=True).update(
    >>>     c=c.item("a") + c.item("b")
    >>> ).into_csv("output.csv")

    >>> list(
    >>>     Table.from_rows(
    >>>         [
    >>>             ("a", "b"),
    >>>             (1, 2),
    >>>             (3, 4),
    >>>         ],
    >>>         header=True,
    >>>     )
    >>>     .update(c=c.col("a") + c.col("b"))
    >>>     .into_iter_rows()
    >>> )
    [("a", "b", "c"), (1, 2, 3), (3, 4, 7)]

    """

    def __init__(
        self,
        row_type: "t.Optional[type]",
        rows_objects: "t.List[t.Iterable]",
        meta_columns: "MetaColumns",
        pending_changes: int,
        pipeline: "t.Optional[BaseConversion]" = None,
        file_to_close=None,
    ):
        """It is used internally only. Use from_rows and from_csv methods."""
        self.row_type = row_type
        self.rows_objects: "t.Optional[t.List[t.Iterable]]" = rows_objects
        self.meta_columns = meta_columns
        self.pending_changes = pending_changes
        self.pipeline = pipeline
        self.file_to_close = file_to_close
        self.row_type = row_type

    def get_columns(self) -> "t.List[str]":
        """Exposes list of column names"""
        return [column.name for column in self.meta_columns.columns]

    columns = property(get_columns)

    @classmethod
    def from_rows(
        cls,
        rows: "t.Iterable[t.Union[dict, tuple, list, t.Any]]",
        header: """t.Optional[
            t.Union[
                bool, t.List[str], t.Tuple[str], t.Dict[str, t.Union[str, int]]
            ]
        ]""" = None,
        # t.Literal["raise", "keep", "drop", "mangle"]
        duplicate_columns: str = "raise",
        skip_rows: int = 0,
        file_to_close=None,
    ) -> "Table":
        """A method to initialize a table conversion from an iterable of
        rows.

        Args:
          rows: can be either an iterable of any objects if no header inference
            is required OR an iterable of dicts, tuples or lists
          header: specifies header inference mode:

             * True: takes either the first tuple/list or keys of the first
               dict as a header
             * False: there's no header in input data, use numbered columns
               instead: COLUMN_0, COLUMN_1, etc.
             * list/tuple of str: there's no header in input data, so this is
               the header to be used (raises ValueError if numbers of columns
               don't match)
             * dict: its keys form the header, values are str/int indices to
               take values of columns from input rows (raises ValueError if
               numbers of columns don't match)
             * None: inspects the first row and if it's a dict, then takes its
               keys as a header

          duplicate_columns: either of following ("raise" by default):

            * "raise": ValueError is raise if a duplicate column is detected
            * "keep": duplicate columns are left as is, but when referenced the
              first one is used
            * "drop": duplicate columns are skipped
            * "mangle": names of duplicate columns are mangled like: "name",
              "name_1", "name_2", etc.

          skip_rows: number of rows to skip at the beginning. Useful when input
            data contains a header, but you provide your own - in this case
            it's convenient to skip the heading row from the input

        """
        columns = MetaColumns(duplicate_columns=duplicate_columns)

        rows = iter(rows)

        if skip_rows:
            for _ in range(skip_rows):
                next(rows)

        first_row = next(rows)
        row_type = type(first_row)
        pending_changes = 0

        index: "t.Union[str, int]"
        if isinstance(header, (tuple, list)):
            if len(header) != len(first_row):
                raise ValueError("non-matching number of columns")
            for index, column in enumerate(header):
                pending_changes |= columns.add(column, index, None)[1]

        elif isinstance(header, dict):
            if len(header) != len(first_row):
                raise ValueError("non-matching number of columns")
            for name, index in header.items():
                pending_changes |= columns.add(name, index, None)[1]

        else:
            # inferring a header
            if isinstance(first_row, dict):
                if header is False:
                    for key in first_row:
                        pending_changes |= columns.add(None, key, None)[1]
                    pending_changes |= ColumnChanges.MUTATE
                else:
                    for key in first_row:
                        pending_changes |= columns.add(key, key, None)[1]

            elif isinstance(first_row, (tuple, list)):
                if header is True:
                    for index, column_name in enumerate(first_row):
                        pending_changes |= columns.add(
                            column_name, index, None
                        )[1]
                    first_row = None

                else:
                    for index in range(len(first_row)):
                        pending_changes |= columns.add(None, index, None)[1]

            else:
                raise ValueError(
                    "failed to infer header: unsupported row type",
                    type(first_row),
                )

        rows_objects: "t.List[t.Iterable]" = [rows]
        if first_row is not None:
            rows_objects.insert(0, (first_row,))

        return cls(
            row_type=row_type,
            rows_objects=rows_objects,
            meta_columns=columns,
            pending_changes=pending_changes,
            file_to_close=file_to_close,
        )

    csv_dialect = CustomCsvDialect

    @classmethod
    def from_csv(
        cls,
        filepath_or_buffer: "t.Union[str, t.TextIO]",
        header: """t.Optional[
            t.Union[
                bool, t.List[str], t.Tuple[str], t.Dict[str, t.Union[str, int]]
            ]
        ]""" = None,
        duplicate_columns: str = "mangle",
        skip_rows: int = 0,
        dialect: "t.Union[str, CustomCsvDialect]" = "excel",
        encoding: str = "utf-8",
    ) -> "Table":
        """A method to initialize a table conversion from a csv-like file.

        Args:
          filepath_or_buffer: a filepath or something :py:obj:`csv.reader` can
            read
          header: specifies header inference mode:

             * True: takes either the first tuple/list or keys of the first
               dict as a header
             * False: there's no header in input data, use numbered columns
               instead: COLUMN_0, COLUMN_1, etc.
             * list/tuple of str: there's no header in input data, so this is
               the header to be used (raises ValueError if numbers of columns
               don't match)
             * dict: its keys form the header, values are str/int indices to
               take values of columns from input rows (raises ValueError if
               numbers of columns don't match)
             * None: inspects the first row and if it's a dict, then takes its
               keys as a header

          duplicate_columns: either of following ("mangle" by default):

            * "raise": ValueError is raise if a duplicate column is detected
            * "keep": duplicate columns are left as is, but when referenced the
              first one is used
            * "drop": duplicate columns are skipped
            * "mangle": names of duplicate columns are mangled like: "name",
              "name_1", "name_2", etc.

          skip_rows: number of rows to skip at the beginning. Useful when input
            data contains a header, but you provide your own - in this case
            it's convenient to skip the heading row from the input

          dialect: a dialect acceptable by :py:obj:`csv.reader` There's a
            helper method:
            :py:obj:`convtools.contrib.tables.Table.csv_dialect` to create
            dialects without defining classes
          encoding: encoding to pass to :py:obj:`open`
        """
        file_to_close = None
        if isinstance(filepath_or_buffer, str):
            file_to_close = open(  # pylint: disable=consider-using-with
                filepath_or_buffer,
                encoding=encoding,
            )
            rows = map(
                tuple,  # type: ignore
                csv.reader(file_to_close, dialect=dialect),
            )
        else:
            rows = map(
                tuple,  # type: ignore
                csv.reader(filepath_or_buffer, dialect=dialect),
            )

        return cls.from_rows(
            rows,
            header,
            duplicate_columns,
            skip_rows=skip_rows,
            file_to_close=file_to_close,
        )

    def embed_conversions(self) -> "Table":
        """There's no need in calling this directly, it's done automatically
        when you use :py:obj:`convtools.contrib.tables.Table.update` or
        :py:obj:`convtools.contrib.tables.Table.join` See the explanation
        below:

        Since each column is either:
          * simply taken by index (cheap)
          * or obtained by performing arbitrary convtools conversion (may be
            expensive)

        it's important to have multiple layers of processing when something
        depends on something which may be expensive to calculate.

        This method adds a new processing stage to a pipeline and exposes all
        "cheap" columns to further conversions.
        """
        if any(column.conversion for column in self.meta_columns.columns):
            column_conversions = []
            for index, column in enumerate(self.meta_columns.columns):
                if column.index is None:
                    column_conversions.append(column.conversion)
                    column.conversion = None
                else:
                    column_conversions.append(GetItem(column.index))
                column.index = index

            conversion = GeneratorComp(tuple(column_conversions))

            self.pipeline = (
                conversion
                if self.pipeline is None
                else self.pipeline.pipe(conversion)
            )
            self.pending_changes = 0
            self.row_type = tuple

        return self

    def filter(self, condition: "BaseConversion") -> "Table":
        """Filters table-like data, keeping rows where ``condition`` resolves
        to ``True``"""
        condition = ensure_conversion(condition)
        column_refs = list(condition.get_dependencies(types=ColumnRef))
        name_to_column = self.meta_columns.get_name_to_column()

        depends_on_complex_columns = any(
            name_to_column[ref.name].conversion is not None
            for ref in column_refs
        )
        if depends_on_complex_columns:
            return self.embed_conversions().filter(condition)

        for ref in condition.get_dependencies(types=ColumnRef):
            column = name_to_column[ref.name]
            ref.set_index(column.index)

        self.pipeline = (self.pipeline or GetItem()).filter(condition)
        return self

    def update(self, **column_to_conversion) -> "Table":
        """The main method to mutate table-like data.

        Args:
          column_to_conversion: dict where keys are new or existing columns and
            values are conversions to be applied row-wise
        """
        column_name_to_column = self.meta_columns.get_name_to_column()

        for column_name in list(column_to_conversion):
            conversion = ensure_conversion(column_to_conversion[column_name])
            column_refs = list(conversion.get_dependencies(types=ColumnRef))

            depends_on_complex_columns = any(
                column_name_to_column[ref.name].conversion is not None
                for ref in column_refs
            )
            if depends_on_complex_columns:
                return self.embed_conversions().update(**column_to_conversion)

            del column_to_conversion[column_name]

            for ref in column_refs:
                ref.set_index(column_name_to_column[ref.name].index)

            if column_name in column_name_to_column:
                column = column_name_to_column[column_name]
                column.conversion = conversion
                column.index = None
            else:
                column, state = self.meta_columns.add(
                    column_name, None, conversion
                )
                self.pending_changes |= state
                column_name_to_column[column.name] = column

            self.pending_changes |= ColumnChanges.MUTATE

        return self

    def update_all(self, *conversions) -> "Table":
        """Table-wide mutations, applied to each value of each column.

        Args:
          conversions: conversion to apply to each value of each column
        """
        conversion: "BaseConversion" = GetItem()
        for conversion_ in conversions:
            conversion = conversion.pipe(conversion_)
        column_to_conversion = {
            column.name: ColumnRef(column.name).pipe(conversion)
            for column in self.meta_columns.columns
        }
        return self.update(**column_to_conversion)

    def rename(
        self, columns: "t.Union[t.Tuple[str], t.List[str], t.Dict[str, str]]"
    ) -> "Table":
        """Method to rename columns. The behavior depends on type of columns
        argument.

        Args:
          columns: if tuple/list, then it defines new column names (length of
            passed columns should match number of columns inside).  If dict,
            then it defines a mapping from old column names to new ones.
        """
        renamed = False
        if isinstance(columns, dict):
            for column_ in self.meta_columns.columns:
                if column_.name in columns:
                    column_.name = columns[column_.name]
                    renamed = True

        elif isinstance(columns, (tuple, list)):
            if len(columns) != len(self.meta_columns.columns):
                raise ValueError("non-matching number of columns")
            for column_, new_column_name in zip(
                self.meta_columns.columns, columns
            ):
                column_.name = new_column_name
            renamed = True

        else:
            raise TypeError("unsupported columns type")

        if renamed and self.row_type is dict:
            self.pending_changes |= ColumnChanges.MUTATE

        return self

    def take(self, *column_names: str) -> "Table":
        """Leaves only specified columns, omitting the rest.

        Args:
          column_names: columns to keep
        """
        self.meta_columns = self.meta_columns.take(*column_names)
        self.pending_changes |= ColumnChanges.REARRANGE
        return self

    def drop(self, *column_names: str) -> "Table":
        """Drops specified columns, keeping the rest.

        Args:
          column_names: columns to drop
        """
        self.meta_columns = self.meta_columns.drop(*column_names)
        if column_names:
            self.pending_changes |= ColumnChanges.REARRANGE
        return self

    def zip(self, table: "Table", fill_value=None) -> "Table":
        """Zip tables one to another. Before using this method, make sure you
        are not looking for :py:obj:`convtools.contrib.tables.Table.join`

        Let's assume fill_value is set to " ":

        >>> Table 1      Table 2
        >>> | a | b |    | b | c |
        >>> | 1 | 2 |    | 3 | 4 |
        >>>              | 5 | 6 |
        >>>
        >>> table1.zip(table2, fill_value=" ")
        >>>
        >>> Result:
        >>> | a | b | b | c |
        >>> | 1 | 2 | 3 | 4 |
        >>> |   |   | 5 | 6 |

        Args:
         - table: table to be chained
         - fill_value: value to use for filling gaps

        """
        new_columns = MetaColumns(duplicate_columns="keep")
        left_columns = self.meta_columns.get_name_to_column()
        right_columns = table.meta_columns.get_name_to_column()

        left_fill_value = tuple(fill_value for _ in range(len(left_columns)))
        right_fill_value = tuple(fill_value for _ in range(len(right_columns)))

        for index, name in enumerate(left_columns):
            new_columns.add(name, None, GetItem(0, index))
        for index, name in enumerate(right_columns):
            new_columns.add(name, None, GetItem(1, index))

        new_rows = (
            (left_fill_value, t[1])
            if t[0] is None
            else ((t[0], right_fill_value) if t[1] is None else t)
            for t in zip_longest(
                self.into_iter_rows(tuple), table.into_iter_rows(tuple)
            )
        )
        return Table(
            row_type=tuple,
            rows_objects=[new_rows],
            meta_columns=new_columns,
            pending_changes=ColumnChanges.MUTATE,
        )

    def chain(
        self,
        table: "Table",
        fill_value=None,
    ) -> "Table":
        """Chain tables, putting them one after another.

        Let's assume fill_value is set to " ":

        >>> Table 1      Table 2
        >>> | a | b |    | b | c |
        >>> | 1 | 2 |    | 3 | 4 |
        >>>
        >>> table1.chain(table2, fill_value=" ")
        >>>
        >>> Result:
        >>> | a | b | c |
        >>> | 1 | 2 |   |
        >>> |   | 3 | 4 |

        Args:
         - table: table to be chained
         - fill_value: value to use for filling gaps

        """
        if self.rows_objects is None:
            raise AssertionError("move_rows called the 2nd time")
        if (
            self.meta_columns.is_same_as(table.meta_columns)
            and not self.pending_changes
            and not self.pipeline
            and not table.pending_changes
            and not table.pipeline
        ):
            self.rows_objects.extend(table.move_rows_objects())
            return self

        first_name_to_columns = self.meta_columns.get_name_to_column()
        second_name_to_columns = table.meta_columns.get_name_to_column()

        new_columns = MetaColumns(duplicate_columns="raise")
        first_columns = MetaColumns(
            duplicate_columns=self.meta_columns.duplicate_columns,
        )
        second_columns = MetaColumns(
            duplicate_columns=table.meta_columns.duplicate_columns,
        )
        fill_value_conversion = NaiveConversion(fill_value)

        index = 0
        for name, first_column in first_name_to_columns.items():
            new_columns.add(name, index, None)
            first_columns.add(*first_column.as_tuple())
            if name in second_name_to_columns:
                second_columns.add(*second_name_to_columns[name].as_tuple())
            else:
                second_columns.add(name, None, fill_value_conversion)
            index += 1

        for name, second_column in second_name_to_columns.items():
            if name in first_name_to_columns:
                continue
            new_columns.add(name, index, None)
            first_columns.add(name, None, fill_value_conversion)
            second_columns.add(*second_column.as_tuple())
            index += 1

        if not self.meta_columns.is_same_as(first_columns):
            self.pending_changes |= ColumnChanges.MUTATE
        self.meta_columns = first_columns

        if not table.meta_columns.is_same_as(second_columns):
            table.pending_changes |= ColumnChanges.MUTATE
        table.meta_columns = second_columns

        row_type = self.row_type or tuple
        rows_objects = (
            self.into_list_of_iterables() + table.into_list_of_iterables()
        )
        return Table(
            row_type=row_type,
            rows_objects=rows_objects,
            meta_columns=new_columns,
            pending_changes=0,
        )

    def join(
        self,
        table: "Table",
        on: "t.Union[BaseConversion, str, t.Iterable[str]]",
        how: str,
        suffixes=("_LEFT", "_RIGHT"),
    ) -> "Table":
        """Joins the table conversion to another table conversion.

        Args:
          table: another table conversion to join to self
          on:
            * either a join conversion like
              ``c.LEFT.col("a") == c.RIGHT.col("A")``
            * or iterable of column names to join on

          how: either of these: "inner", "left", "right", "outer" (same as
            "full")
          suffixes: tuple of two strings: the first one is the suffix to be
            added to left columns, having conflicting names with right columns;
            the second one is added to conflicting right ones. When ``on`` is
            an iterable of strings, these columns are excluded from suffixing.
        """

        how = JoinConversion.validate_how(how)
        left = self.embed_conversions()
        right = table.embed_conversions()
        left_join_conversion = LeftJoinCondition()
        right_join_conversion = RightJoinCondition()
        left_column_name_to_column = left.meta_columns.get_name_to_column()
        right_column_name_to_column = right.meta_columns.get_name_to_column()

        after_join_conversions: "t.List[BaseConversion]" = []
        after_join_column_names: "t.List[str]" = []

        if isinstance(on, BaseConversion):
            # intentionally left blank to force suffixing
            join_columns = set()
            join_condition = on
            for ref in join_condition.get_dependencies(types=ColumnRef):
                if ref.id_ == left_join_conversion.NAME:
                    column = left_column_name_to_column[ref.name]
                elif ref.id_ == right_join_conversion.NAME:
                    column = right_column_name_to_column[ref.name]
                else:
                    raise ValueError("ambiguous column", ref.name)
                ref.set_index(column.index)
        else:
            on = [on] if isinstance(on, str) else list(on)
            join_columns = set(on)
            join_condition = (
                And(
                    *(
                        left_join_conversion.item(
                            left_column_name_to_column[column_name].index
                        )
                        == right_join_conversion.item(
                            right_column_name_to_column[column_name].index
                        )
                        for column_name in on
                    )
                )
                if len(on) > 1
                else left_join_conversion.item(
                    left_column_name_to_column[on[0]].index
                )
                == right_join_conversion.item(
                    right_column_name_to_column[on[0]].index
                )
            )
        del on

        only_left_values_matter = how in ("left", "inner")
        left_is_optional = how in ("right", "outer")
        right_is_optional = how in ("left", "outer")
        for column in left.meta_columns.columns:
            index = column.index
            column_name = column.name
            if column_name in right_column_name_to_column:
                if column_name in join_columns:
                    after_join_column_names.append(column_name)
                    if only_left_values_matter:
                        after_join_conversions.append(GetItem(0, index))
                    elif how == "right":
                        after_join_conversions.append(GetItem(1, index))
                    else:  # outer
                        after_join_conversions.append(
                            If(
                                GetItem(0).is_(None),
                                GetItem(1, index),
                                GetItem(0, index),
                            )
                        )
                else:
                    after_join_column_names.append(
                        f"{column_name}{suffixes[0]}"
                    )
                    after_join_conversions.append(
                        If(GetItem(0).is_(None), None, GetItem(0, index))
                        if left_is_optional
                        else GetItem(0, index)
                    )
            else:
                after_join_column_names.append(column_name)
                after_join_conversions.append(
                    If(GetItem(0).is_(None), None, GetItem(0, index))
                    if left_is_optional
                    else GetItem(0, index)
                )

        for column in right.meta_columns.columns:
            index = column.index
            column_name = column.name
            if column_name in left_column_name_to_column:
                if column_name in join_columns:
                    # handled above
                    pass
                else:
                    after_join_column_names.append(
                        f"{column_name}{suffixes[1]}"
                    )
                    after_join_conversions.append(
                        If(GetItem(1).is_(None), None, GetItem(1, index))
                        if right_is_optional
                        else GetItem(1, index)
                    )
            else:
                after_join_column_names.append(column_name)
                after_join_conversions.append(
                    If(GetItem(1).is_(None), None, GetItem(1, index))
                    if right_is_optional
                    else GetItem(1, index)
                )

        new_rows = JoinConversion(
            GetItem(),
            InputArg("right"),
            join_condition,
            how,
        ).execute(
            left.into_iter_rows(left.row_type),
            right=right.into_iter_rows(left.row_type),
            debug=ConverterOptionsCtx.get_option_value("debug"),
        )
        new_columns = MetaColumns(
            duplicate_columns="raise",
        )
        for column_name, conversion in zip(
            after_join_column_names, after_join_conversions
        ):
            new_columns.add(column_name, None, conversion)

        return Table(
            row_type=tuple,
            rows_objects=[new_rows],
            meta_columns=new_columns,
            pending_changes=ColumnChanges.MUTATE,
        )

    def move_rows_objects(self) -> "t.List[t.Iterable]":
        """Moves out rows objects including files to be closed later"""
        if self.rows_objects is None:
            raise AssertionError("move_rows called the 2nd time")

        rows_objects = self.rows_objects
        self.rows_objects = None

        if self.file_to_close:
            rows_objects.append(CloseFileIterator(self.file_to_close))
            self.file_to_close = None
        return rows_objects

    supported_types = (tuple, list, dict)

    def into_list_of_iterables(
        self, type_=tuple, include_header=None
    ) -> "t.List[t.Iterable]":
        if type_ not in self.supported_types:
            raise TypeError("unsupported type_", type_)

        no_pending_changes = (
            type_ is self.row_type and not self.pending_changes
        )

        if no_pending_changes:
            conversion = None if self.pipeline is None else self.pipeline
        else:
            row_conversion: "t.Union[dict, tuple, list]"
            if type_ is dict:
                row_conversion = {
                    column.name: (
                        column.conversion
                        if column.index is None
                        else GetItem(column.index)
                    )
                    for column in self.meta_columns.columns
                }
                include_header = False
            else:
                row_conversion = type_(
                    (
                        column.conversion
                        if column.index is None
                        else GetItem(column.index)
                    )
                    for column in self.meta_columns.columns
                )
            conversion = (self.pipeline or GetItem()).pipe(
                GeneratorComp(row_conversion)
            )

        if conversion:
            converter = conversion.gen_converter()
            rows_objects = [
                converter(chunk) for chunk in self.move_rows_objects()
            ]
        else:
            rows_objects = self.move_rows_objects()

        if type_ is not dict and include_header:
            header = type_(self.get_columns())
            rows_objects.insert(0, (header,))

        return rows_objects

    def into_iter_rows(
        self, type_=tuple, include_header=None
    ) -> "t.Iterable[t.Any]":
        """Consumes inner rows Iterable and returns Iterable of processed rows.

        Args:
          type_: casts output rows to the type. Accepts the following values:

            * :py:obj:`dict`
            * :py:obj:`tuple`
            * :py:obj:`list`

        """
        return chain.from_iterable(
            self.into_list_of_iterables(
                type_=type_, include_header=include_header
            )
        )

    def into_csv(
        self,
        filepath_or_buffer: "t.Union[str, t.TextIO]",
        include_header: bool = True,
        dialect: "t.Union[str, CustomCsvDialect]" = "excel",
        encoding="utf-8",
    ):
        """Consumes inner rows Iterable and writes processed rows as a csv-like
        file.

        Args:
          filepath_or_buffer: a filepath or something :py:obj:`csv.writer` can
            write to
          include_header: self-explanatory bool
          dialect: a dialect acceptable by :py:obj:`csv.writer` There's a
            helper method:
            :py:obj:`convtools.contrib.tables.Table.csv_dialect` to create
            dialects without defining classes
          encoding: encoding to pass to :py:obj:`open`
        """
        row_type = list if self.row_type is list else tuple

        f_to_close = None
        if isinstance(filepath_or_buffer, str):
            f = f_to_close = open(  # pylint:disable=consider-using-with
                filepath_or_buffer, "w", encoding=encoding
            )
        else:
            f = filepath_or_buffer

        try:
            writer = csv.writer(f, dialect=dialect)
            if include_header:
                writer.writerow(self.get_columns())
            for chunk in self.into_list_of_iterables(
                row_type, include_header=False
            ):
                writer.writerows(chunk)
        finally:
            if f_to_close is not None:
                f_to_close.close()



"""
Provides conversions for slicing iterables into chunks.
"""
import typing as t

from .aggregations import Aggregate
from .base import (
    BaseConversion,
    Code,
    ConversionWrapper,
    EscapedString,
    GeneratorComp,
    NamedConversion,
    This,
)


class BaseChunkBy(BaseConversion):
    def aggregate(self, *args, **kwargs) -> "BaseConversion":
        return self.pipe(GeneratorComp(Aggregate(*args, **kwargs)))


class ChunkBy(BaseChunkBy):
    """Slices iterable into chunks by element values and/or size of chunks.

    >>> # simple #1
    >>> c.chunk_by(size=1000)
    >>>
    >>> # simple #2
    >>> c.chunk_by(c.item("x"))
    >>>
    >>> # with aggregate
    >>> c.chunk_by(
    >>>     c.item("x"),
    >>>     size=1000
    >>> ).aggregate({
    >>>     "x": c.ReduceFuncs.Last(c.item("x")),
    >>>     "y": c.ReduceFuncs.Sum(c.item("y")),
    >>> })


    >>> # example #1
    >>> c.chunk_by_condition(c.CHUNK.len() < 100)
    >>>
    >>> # example #2
    >>> c.chunk_by_condition(
    >>>     c.and_(
    >>>         c.CHUNK.item(-1) == c.this(),
    >>>         c.CHUNK.item(-1) - c.this() < 100
    >>>     )
    >>> )
    >>> # with aggregate
    >>> c.chunk_by_condition(
    >>>     c.CHUNK.len() < 100
    >>> ).aggregate({
    >>>     "x": c.ReduceFuncs.Last(c.item("x")),
    >>>     "y": c.ReduceFuncs.Sum(c.item("y")),
    >>> })

    It also provides a shortcut for running
    :py:obj:`convtools.aggregations.Aggregate` on chunks.
    """

    def __init__(self, *by, size: t.Optional[int] = None):
        """
        Args:
          by: fields/conversions to use for slicing into chunks (elements with
            same values go to the same chunk)
          size: (optional) positive int to limit max size of a chunk
        """
        super().__init__()
        if size is not None:
            if not isinstance(size, int) or size <= 0:
                raise ValueError("size has to be positive int or None")
        if not by and not size:
            raise ValueError("pass at least one of by or size params")

        self.by = (
            (self.ensure_conversion(by if len(by) > 1 else by[0]))
            if by
            else None
        )
        self.size = size

    def _gen_code_and_update_ctx(self, code_input, ctx):
        converter_name = self.gen_name("chunk_by", ctx, self)
        code_args = self.by.get_args_def_code() if self.by else ""
        func_args = self.by.get_args_as_func_args() if self.by else ()

        code = Code()
        code.add_line(f"def {converter_name}(items_{code_args}):", 1)
        code.add_line("items_ = iter(items_)", 0)
        code.add_line("try:", 0)
        code.add_line("    item_ = next(items_)", 0)
        code.add_line("except StopIteration:", 0)
        code.add_line("    return", 0)

        code.add_line("chunk_ = [item_]", 0)

        code_item_to_signature = (
            self.by.gen_code_and_update_ctx("item_", ctx) if self.by else None
        )

        code_before_for = Code()
        code_after_for = Code()
        code_if_condition = None
        code_if_continue_chunk = Code()
        code_if_new_chunk = Code()

        code_if_continue_chunk.add_line("chunk_.append(item_)", 0)
        code_if_new_chunk.add_line("yield chunk_", 0)
        code_if_new_chunk.add_line("chunk_ = [item_]", 0)

        if code_item_to_signature and self.size:
            code_if_condition = (
                "chunk_item_signature == new_item_signature and size_ < "
                f"{self.size}"
            )

        if code_item_to_signature:
            code_before_for.add_line(
                f"chunk_item_signature = {code_item_to_signature}", 0
            )
            code_after_for.add_line(
                f"new_item_signature = {code_item_to_signature}", 0
            )
            code_if_condition = (
                code_if_condition
                or "chunk_item_signature == new_item_signature"
            )
            code_if_new_chunk.add_line(
                "chunk_item_signature = new_item_signature", 0
            )

        if self.size:
            code_before_for.add_line("size_ = 1", 0)
            code_if_condition = code_if_condition or f"size_ < {self.size}"
            code_if_continue_chunk.add_line("size_ = size_ + 1", 0)
            code_if_new_chunk.add_line("size_ = 1", 0)

        if not code_if_condition:
            raise AssertionError("impossible case")

        code.add_code(code_before_for)
        code.add_line("for item_ in items_:", 1)
        code.add_code(code_after_for)
        code.add_line(f"if {code_if_condition}:", 1)
        code.add_code(code_if_continue_chunk)
        code.incr_indent_level(-1)
        code.add_line("else:", 1)
        code.add_code(code_if_new_chunk)
        code.incr_indent_level(-2)
        code.add_line("yield chunk_", -1)

        self._code_to_converter(converter_name, code.to_string(0), ctx)

        return (
            EscapedString(converter_name)
            .call(This(), *func_args)
            .gen_code_and_update_ctx(code_input, ctx)
        )


CHUNK_BY_CONDITION_TEMPLATE = """
def {converter_name}(items_{code_args}):
    items_ = iter(items_)
    try:
        chunk_ = [next(items_)]
    except StopIteration:
        return

    for item_ in items_:
        if {code_condition}:
            chunk_.append(item_)
        else:
            yield chunk_
            chunk_ = [item_]

    yield chunk_
"""


class ChunkByCondition(BaseChunkBy):
    """Slices iterable into chunks based on condition, which is a function of a
    current chunk and a current element.

    >>> # example #1
    >>> c.chunk_by_condition(c.CHUNK.len() < 100)
    >>>
    >>> # example #2
    >>> c.chunk_by_condition(
    >>>     c.and_(
    >>>         c.CHUNK.item(-1) == c.this(),
    >>>         c.CHUNK.item(-1) - c.this() < 100
    >>>     )
    >>> )
    >>> # with aggregate
    >>> c.chunk_by_condition(
    >>>     c.CHUNK.len() < 100
    >>> ).aggregate({
    >>>     "x": c.ReduceFuncs.Last(c.item("x")),
    >>>     "y": c.ReduceFuncs.Sum(c.item("y")),
    >>> })

    It also provides a shortcut for running
    :py:obj:`convtools.aggregations.Aggregate` on chunks.
    """

    CHUNK = NamedConversion("chunk_", This())

    def __init__(self, condition):
        super().__init__()
        self.condition = ConversionWrapper(
            self.ensure_conversion(condition),
            name_to_code_input={self.CHUNK.name: "chunk_"},
        )

    def _gen_code_and_update_ctx(self, code_input, ctx):
        converter_name = self.gen_name("chunk_by_condition", ctx, self)
        code = CHUNK_BY_CONDITION_TEMPLATE.format(
            converter_name=converter_name,
            code_args=self.condition.get_args_def_code(),
            code_condition=self.condition.gen_code_and_update_ctx(
                "item_", ctx
            ),
        )
        self._code_to_converter(converter_name, code, ctx)
        return (
            EscapedString(converter_name)
            .call(
                This(),
                *self.condition.get_args_as_func_args(),
            )
            .gen_code_and_update_ctx(code_input, ctx)
        )



"""
Implements base conversions to reference and define columns in table
conversions
"""
import typing as t
from collections import defaultdict

from .base import BaseConversion, ConversionException, GetItem


class ColumnRef(BaseConversion):
    """Helper conversion, which allows to reference a column of a table
    conversion"""

    def __init__(self, name: str, id_=None):
        super().__init__()
        if not isinstance(name, str):
            raise ValueError("name should be str")
        self.name = name
        self.index: "t.Optional[t.Union[str, int]]" = None
        self.id_ = id_

    def set_index(self, index: t.Union[str, int]):
        if not isinstance(index, (str, int)):
            raise AssertionError("bad index")
        self.index = index
        return self

    def _gen_code_and_update_ctx(self, code_input, ctx):
        if self.index is None:
            raise ConversionException(
                "column index is not initialized, "
                "possible use outside of Table"
            )
        return GetItem(
            self.index if self.index is not None else -1
        ).gen_code_and_update_ctx(code_input, ctx)


class ColumnDef:
    """
    Defines a column within a table conversion. It holds the following:

     * name of the column in the output
     * index of the column in the input in simple cases, otherwise None
     * conversion to obtain the value from the input, None in simple cases

    """

    __slots__ = ["name", "index", "conversion"]

    def __init__(
        self,
        name: str,
        index: t.Optional[t.Any],
        conversion: t.Optional[BaseConversion],
    ):
        if not bool(index is not None) ^ bool(conversion is not None):
            raise ValueError("provide either index or conversion")
        self.name = name
        self.index = index
        self.conversion = conversion

    def as_tuple(self):
        return self.name, self.index, self.conversion

    def is_same_as(self, other: "ColumnDef") -> bool:
        return (
            self.name == other.name
            and self.index == other.index
            and self.conversion is None
            and other.conversion is None
        )


class ColumnChanges:
    RENAME = 1
    REARRANGE = 2
    MUTATE = 4


class MetaColumns:
    """A helper container for naming & keeping column definitions"""

    def __init__(
        self,
        # t.Literal["raise", "keep", "drop", "mangle"]
        duplicate_columns="raise",
    ):
        self.columns: "t.List[ColumnDef]" = []
        self.column_to_number = defaultdict(int)
        if duplicate_columns not in ("raise", "keep", "drop", "mangle"):
            raise ValueError("invalid duplicate_columns value")
        self.duplicate_columns = duplicate_columns

    def is_same_as(self, other: "MetaColumns"):
        left_columns = self.columns
        right_columns = other.columns
        if len(left_columns) != len(right_columns):
            return False

        for index in range(len(left_columns)):
            if not left_columns[index].is_same_as(right_columns[index]):
                return False

        return True

    def add(self, name, index, conversion):
        column_number = self.column_to_number[name]
        self.column_to_number[name] += 1
        state = 0

        if name is None:
            name = f"COLUMN_{column_number}"
        elif column_number:
            if self.duplicate_columns == "raise":
                raise ValueError("such column already exists", name)
            if self.duplicate_columns == "mangle":
                name = f"{name}_{column_number}"
                state = ColumnChanges.RENAME
            elif self.duplicate_columns == "drop":
                return None, ColumnChanges.REARRANGE

        column = ColumnDef(name, index, conversion)
        self.columns.append(column)
        return column, state

    def take(self, *column_names: str) -> "MetaColumns":
        name_to_column = self.get_name_to_column()
        missing_columns = set(column_names).difference(name_to_column)
        if missing_columns:
            raise ValueError("missing columns", missing_columns)

        new_columns = MetaColumns(self.duplicate_columns)

        for column_name in column_names:
            column = name_to_column[column_name]
            new_columns.add(column.name, column.index, column.conversion)

        return new_columns

    def drop(self, *column_names: str) -> "MetaColumns":
        unique_column_names = set(column_names)
        missing_columns = unique_column_names.difference(
            {column.name for column in self.columns}
        )
        if missing_columns:
            raise ValueError("missing columns", missing_columns)

        new_columns = MetaColumns(self.duplicate_columns)
        for column in self.columns:
            if column.name in unique_column_names:
                continue
            new_columns.add(column.name, column.index, column.conversion)

        return new_columns

    def get_name_to_column(self) -> "t.Dict[str, ColumnDef]":
        return {column.name: column for column in self.columns}
